import errno
import os
from pathlib import Path
from typing import Iterator, NewType

from fuse import FuseOSError, Operations, fuse_get_context

from ..core.utils import full_path, log, on_error

ModeT = NewType("ModeT", int)
FileHandleT = NewType("FileHandleT", int)
FileDescriptorT = NewType("FileDescriptorT", int)


class Passthrough(Operations):
    def __init__(self, root: str):
        self.root = Path(root)

    def _full_path(self, partial) -> Path:
        return self.root / partial.lstrip("/")

    @full_path
    @log
    def access(self, path: str, mode: ModeT):
        if not os.access(path, mode):
            raise FuseOSError(errno.EACCES)

    @full_path
    @log
    def chmod(self, path: str, mode: ModeT) -> None:
        try:
            os.chmod(path, mode)
        except OSError as exc:
            on_error(exc)

    @full_path
    @log
    def chown(self, path: str, uid: int, gid: int) -> None:
        try:
            os.chown(path, uid, gid)
        except OSError as exc:
            on_error(exc)

    @full_path
    @log
    def getattr(self, path: str, fh=None) -> dict:
        st = os.lstat(path)
        return dict(
            (key, getattr(st, key))
            for key in (
                "st_atime",
                "st_ctime",
                "st_gid",
                "st_mode",
                "st_mtime",
                "st_nlink",
                "st_size",
                "st_uid",
            )
        )

    @full_path
    @log
    def readdir(self, path: str, fh) -> Iterator[str]:
        dirents = [".", ".."]
        if os.path.isdir(path):
            dirents.extend(os.listdir(path))
        for r in dirents:
            yield r

    @full_path
    @log
    def readlink(self, path) -> str:
        try:
            pathname = os.readlink(path)
        except OSError as exc:
            on_error(exc)
        else:
            if pathname.startswith("/"):
                return os.path.relpath(pathname, self.root)
            else:
                return pathname

    @full_path
    @log
    def mknod(self, path: str, mode: ModeT, dev) -> None:
        try:
            os.mknod(path, mode, dev)
        except OSError as exc:
            on_error(exc)

    @full_path
    @log
    def rmdir(self, path) -> None:
        try:
            os.rmdir(path)
        except FileNotFoundError as exc:
            on_error(exc)
        except OSError as exc:
            on_error(exc)

    @full_path
    @log
    def mkdir(self, path: str, mode: ModeT) -> None:
        try:
            os.mkdir(path, mode)
        except FileExistsError as exc:
            on_error(exc)
        except FileNotFoundError as exc:
            on_error(exc)
        except OSError as exc:
            on_error(exc)

    @full_path
    @log
    def statfs(self, path: str) -> dict:
        try:
            stv = os.statvfs(path)
        except OSError as exc:
            on_error(exc)
        else:
            return dict(
                (key, getattr(stv, key))
                for key in (
                    "f_bavail",
                    "f_bfree",
                    "f_blocks",
                    "f_bsize",
                    "f_favail",
                    "f_ffree",
                    "f_files",
                    "f_flag",
                    "f_frsize",
                    "f_namemax",
                )
            )

    @full_path
    @log
    def unlink(self, path) -> None:
        try:
            os.unlink(path)
        except FileNotFoundError as exc:
            on_error(exc)
        except OSError as exc:
            on_error(exc)

    @log
    def symlink(self, name, target, *args) -> None:
        path = self._full_path(name)
        uid, gid, pid = fuse_get_context()
        try:
            os.symlink(target, path)
            os.chown(path, uid, gid, follow_symlinks=False)
        except OSError as exc:
            on_error(exc)

    @log
    def rename(self, old: str, new: str) -> None:
        try:
            os.rename(self._full_path(old), self._full_path(new))
        except FileExistsError as exc:
            on_error(exc)
        except IsADirectoryError as exc:
            on_error(exc)
        except NotADirectoryError as exc:
            on_error(exc)
        except OSError as exc:
            on_error(exc)

    @log
    def link(self, target, name) -> None:
        try:
            os.link(self._full_path(name), self._full_path(target))
        except OSError as exc:
            on_error(exc)

    @full_path
    @log
    def utimens(self, path: str, times=None) -> None:
        try:
            os.utime(path, times)
        except OSError as exc:
            on_error(exc)

    @full_path
    @log
    def open(self, path: str, flags) -> FileDescriptorT:
        try:
            return FileDescriptorT(os.open(path, flags))
        except OSError as exc:
            on_error(exc)

    @full_path
    @log
    def create(self, path: str, mode: ModeT, fi=None) -> FileDescriptorT:
        uid, gid, pid = fuse_get_context()
        try:
            fd = os.open(path, os.O_WRONLY | os.O_CREAT, mode)
            os.chown(path, uid, gid)
        except OSError as exc:
            on_error(exc)
        else:
            return FileDescriptorT(fd)

    @log
    def read(self, path: str, length, offset, fh) -> bytes:
        os.lseek(fh, offset, os.SEEK_SET)
        return os.read(fh, length)

    @log
    def write(self, path: str, buf, offset, fh) -> int:
        os.lseek(fh, offset, os.SEEK_SET)
        return os.write(fh, buf)

    @full_path
    @log
    def truncate(self, path: str, length, fh: FileHandleT = None) -> None:
        try:
            with open(path, "r+") as f:
                f.truncate(length)
        except OSError as exc:
            on_error(exc)

    @log
    def flush(self, path: str, fh) -> None:
        try:
            os.fsync(fh)
        except OSError as exc:
            on_error(exc)

    @log
    def release(self, path: str, fh: FileHandleT) -> None:
        try:
            os.close(fh)
        except OSError as exc:
            on_error(exc)

    @full_path
    @log
    def fsync(self, path: str, fdatasync, fh: FileHandleT) -> None:
        self.flush(path, fh)



from pydantic_settings import BaseSettings, SettingsConfigDict


class Settings(BaseSettings):
    S3_API: str
    S3_ACCESS_KEY: str
    S3_SECRET_KEY: str

    model_config = SettingsConfigDict(env_file=".env", env_file_encoding="utf-8")


settings = Settings()



from collections.abc import AsyncGenerator
from contextlib import asynccontextmanager
from typing import Any

import aioboto3
from boto3_type_annotations.s3 import Client

from .config import settings

session = aioboto3.Session(
    aws_access_key_id=settings.S3_ACCESS_KEY,
    aws_secret_access_key=settings.S3_SECRET_KEY,
)


@asynccontextmanager
async def get_s3_client() -> AsyncGenerator[Client, Any]:
    client: Client = session.client("s3", endpoint_url=settings.S3_API)
    try:
        yield client
    finally:
        await client.close()



import logging
import os
import queue
from datetime import datetime
from logging import LogRecord
from logging.config import dictConfig
from logging.handlers import QueueHandler, QueueListener, RotatingFileHandler
from typing import cast

import rich.logging
import structlog
from rich.console import Console, ConsoleRenderable, RenderableType
from rich.containers import Renderables
from rich.table import Table
from rich.text import Text, TextType
from rich.traceback import Traceback

from .. import pys3fuse_dir

pys3fuse_logs_dir = pys3fuse_dir / "logs"

if not pys3fuse_logs_dir.exists():
    os.mkdir(pys3fuse_logs_dir, 0o766)

cons = Console(force_terminal=True)


class DefaultLogRender:
    def __init__(self, time_format="[%x %X]") -> None:
        self.time_format = time_format
        self.omit_repeated_times = False
        self.level_width = None
        self.last_func_name = None

    def __call__(
        self,
        console: Console,
        renderables: list[ConsoleRenderable],
        log_time: datetime,
        level: TextType,
        func_name: str,
    ) -> Table:
        output = Table.grid(padding=(0, 1))
        output.expand = True
        output.add_column(style="log.time")
        output.add_column(style="log.level")
        output.add_column(style="magenta1 bold")
        output.add_column(ratio=1, style="log.message", overflow="fold")

        row: list[RenderableType] = []

        log_time = log_time or console.get_datetime()
        log_time_display = Text(log_time.strftime(self.time_format))
        row.append(log_time_display)
        row.append(level)

        if self.last_func_name is None:
            func_name_text = Text()
            func_name_text.append(f"[{func_name:<8}]")
            self.last_func_name = func_name
        elif self.last_func_name == func_name:
            func_name_text = "          "
        else:
            func_name_text = Text()
            func_name_text.append(f"[{func_name:<8}]")
            self.last_func_name = None
        row.append(func_name_text)
        row.append(Renderables(renderables))
        output.add_row(*row)
        return output


class PyS3FUSERichHandler(rich.logging.RichHandler):
    def __init__(self):
        super().__init__(
            console=cons,
            markup=True,
            rich_tracebacks=True,
            tracebacks_show_locals=True,
        )
        self._log_render = DefaultLogRender("%Y-%m-%d %H:%M:%S")

    def render(
        self,
        *,
        record: logging.LogRecord,
        traceback: Traceback | None,
        message_renderable: "ConsoleRenderable",
    ) -> "ConsoleRenderable":
        level = self.get_level_text(record)
        log_time = datetime.fromtimestamp(record.created)

        func_name = getattr(record, "func_name", None)
        if func_name is None:
            func_name = f"{record.name}::{record.filename}::{record.funcName}"

        log_renderable = self._log_render(
            self.console,
            [message_renderable] if not traceback else [message_renderable, traceback],
            log_time,
            level,
            func_name,
        )
        return log_renderable


class PyS3FUSEQueueListener(QueueListener):
    def handle(self, record: LogRecord):
        record = self.prepare(record)
        handler = logging.getHandlerByName("console_handler")

        if not self.respect_handler_level:
            process = True
        else:
            process = record.levelno >= handler.level
        if process:
            handler.handle(record)


class PyS3FUSEQueueHandler(QueueHandler):
    listener: PyS3FUSEQueueListener

    def prepare(self, record: LogRecord) -> LogRecord:
        return record


class PyS3FUSEStructLogProcessor:
    def __call__(self, _, __, event_dict: dict):
        log_dict = {
            "timestamp": event_dict["timestamp"],
            "logger": event_dict["logger"],
            "level": event_dict["level"],
            "message": event_dict["message"],
        }
        if (exception := event_dict.get("exception", None)) is not None:
            log_dict["exception"] = exception
        return log_dict


dictConfig(
    {
        "version": 1,
        "disable_existing_loggers": False,
        "loggers": {
            "pys3fuse": {
                "level": logging.DEBUG,
                "handlers": ["queue_handler", "file_queue_handler"],
            },
            "fuse": {
                "level": logging.CRITICAL,
                "handlers": ["queue_handler", "file_queue_handler"],
            },
        },
        "handlers": {
            "console_handler": {"class": PyS3FUSERichHandler},
            "file_handler": {
                "class": RotatingFileHandler,
                "formatter": "file_formatter",
                "filename": pys3fuse_dir / "logs" / "pys3fuse.log",
                "maxBytes": 10 * 1024 * 1024,  # 10 MB
                "backupCount": 10,
                "encoding": "utf8",
            },
            "queue_handler": {
                "class": PyS3FUSEQueueHandler,
                "queue": queue.Queue(),
                "listener": PyS3FUSEQueueListener,
                "handlers": ["console_handler"],
            },
            "file_queue_handler": {
                "class": PyS3FUSEQueueHandler,
                "queue": queue.Queue(),
                "listener": QueueListener,
                "handlers": ["file_handler"],
            },
        },
        "formatters": {
            "file_formatter": {
                "()": structlog.stdlib.ProcessorFormatter,
                "processors": [
                    structlog.stdlib.add_log_level,
                    structlog.stdlib.add_logger_name,
                    structlog.processors.TimeStamper(fmt="%Y-%m-%d %H:%M:%S"),
                    structlog.processors.ExceptionRenderer(
                        structlog.processors.ExceptionDictTransformer()
                    ),
                    structlog.processors.EventRenamer("message"),
                    PyS3FUSEStructLogProcessor(),
                    structlog.processors.LogfmtRenderer(),
                ],
            },
        },
    }
)

pys3fuse_logger = logging.getLogger("pys3fuse")
queue_listener: PyS3FUSEQueueListener = cast(
    PyS3FUSEQueueHandler, logging.getHandlerByName("queue_handler")
).listener
file_queue_listener: QueueListener = cast(  # noqa
    QueueHandler, logging.getHandlerByName("file_queue_handler")
).listener



import errno
import inspect
from functools import singledispatch, wraps
from typing import Never, ParamSpec

from fuse import FuseOSError

from .log_config import pys3fuse_logger


@singledispatch
def on_error(exc: Exception, *, errno_: int = 0) -> Never:
    """Log Exception and raise `FuseOSError`"""
    raise FuseOSError(errno_)


@on_error.register
def on_os_error(exc: OSError) -> Never:
    raise FuseOSError(exc.errno)


@on_error.register
def on_key_error(exc: KeyError) -> Never:
    raise FuseOSError(errno.ENOENT)


@on_error.register
def on_is_a_dir_error(exc: IsADirectoryError) -> Never:
    raise FuseOSError(exc.errno)


@on_error.register
def on_not_a_dir_error(exc: NotADirectoryError) -> Never:
    raise FuseOSError(exc.errno)


@on_error.register
def on_file_not_found_error(exc: FileNotFoundError) -> Never:
    raise FuseOSError(exc.errno)


@on_error.register
def on_file_exists(exc: FileExistsError) -> Never:
    raise FuseOSError(exc.errno)


def log(func):
    @wraps(func)
    def inner(*args: P.args, **kwargs: P.kwargs):
        func_sig = inspect.signature(func)
        params = dict(zip(func_sig.parameters, args))
        del params["self"]

        log_msg = ", ".join(f"{k}={v}" for k, v in params.items())
        pys3fuse_logger.debug(log_msg, extra={"func_name": func.__name__})
        return func(*args, **kwargs)

    return inner


P = ParamSpec("P")


def full_path(func):
    def inner(*args: P.args, **kwargs: P.kwargs):
        func_sig = inspect.signature(func)
        params = dict(zip(func_sig.parameters, args))

        self = params["self"]
        params["path"] = self.root / params["path"].lstrip("/")

        args = tuple(params.values())
        return func(*args, **kwargs)

    return inner



import logging
from argparse import ArgumentParser

from fuse import FUSE

from .core.log_config import (
    file_queue_listener,
    pys3fuse_logger as logger,
    queue_listener,
)
from .fs.passthrough import Passthrough

if __name__ == "__main__":
    queue_listener.start()
    file_queue_listener.start()

    parser = ArgumentParser("PyS3FUSE")

    parser.add_argument(
        "source",
        type=str,
        help="Directory tree to mirror",
    )
    parser.add_argument(
        "mountpoint",
        type=str,
        help="Where to mount the file system",
    )
    parser.add_argument(
        "--debug",
        action="store_true",
        default=False,
        help="Enable debugging output",
    )

    args = parser.parse_args()
    if args.debug:
        logger.setLevel(logging.DEBUG)

    try:
        FUSE(
            Passthrough(args.source),
            args.mountpoint,
            foreground=True,
            allow_other=True,
        )
    finally:
        queue_listener.stop()
        file_queue_listener.start()
